﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Text;
using System.Windows.Forms;
using System.Drawing;
using System.IO;
using System.Runtime.InteropServices;

using mshtml;
using MsHtmHstInterop;
using SHDocVw;

using MapApp.Model;

namespace MapApp.Ui.Controls
{

    public partial class GoogleMapControl : System.Windows.Forms.WebBrowser
    {
        private string _helpNamespace;
        private string _helpTopic;
        private MapLocation _initialCenterAndZoom =
                new MapLocation("", -31.94, 115.83, 3);
        private int _initialMapType = 0;

        private static string INIT_CENTER_AND_ZOOM_TOKEN = "//<!--SetInitialCenterAndZoom-->";
        private static string INIT_MAP_TYPE_TOKEN = "//<!--SetInitialMapType-->";
        ////private static string INIT_MARKERS_TOKEN = "//<!--InitialMarkersToken-->";

        private ZoomHistory _zoomHistory = new ZoomHistory();



        private static string mapFilePath = AppUtil.TempHtmlDir + "/map.html";
        private static string pdmarkerFilePath = AppUtil.TempHtmlDir + "/pdmarker.js";
        private static string cssFilePath = AppUtil.TempHtmlDir + "/style.css";

        public delegate void OnMapMouseMove(GoogleMapControl sender, MapLocation loc);
        public event OnMapMouseMove MapMouseMove;

        public delegate void OnMapClick(GoogleMapControl sender, MapClickEventArgs e);
        public event OnMapClick MapClick;

        public delegate void OnShowStreetView(GoogleMapControl sender, MapLocation loc);
        public event OnShowStreetView ShowStreetView;

        private Map _map = new Map();
        private MapLocation _selectedMarker;

        private bool _showAllLabelsState = false;


        #region Constructors
        public GoogleMapControl()
        {
            InitializeComponentAndCustomInitialize();
        }

        public GoogleMapControl(IContainer container)
        {
            container.Add(this);

            InitializeComponentAndCustomInitialize();
        }

        /// <summary>
        /// Used in constructors
        /// </summary>
        private void InitializeComponentAndCustomInitialize()
        {
            InitializeComponent();
            CustomInitialize();
        }

        private void CustomInitialize()
        {
            this.ScrollBarsEnabled = false;

            // Copy the stylesheet to temp directory:
            AppUtil.WriteResourceToDisk("MapApp.Ui.Controls.Resources.html.style.css", cssFilePath);
            //// Copy the map.html & pdmarker.js files from resources to temp dir:


            AppUtil.WriteResourceToDisk("MapApp.Ui.Controls.Resources.html.pdmarker.js", pdmarkerFilePath);

            if (_map != null)
            {
                _map.MapLocations.RaiseListChangedEvents = true;
                _map.MapLocations.ListChanged += new ListChangedEventHandler(MapLocations_ListChanged);
                _map.MapLocationsChanged += new Map.MapLocationsChangedEventHandler(_map_MapLocationsChanged);
                DrawMapLocationCollection(_map.MapLocations);
            }

            this.MapClick += new OnMapClick(GoogleMapControl_MapClick);

            newMarkerContextMenuItem.Click += new EventHandler(newMarkerContextMenuItem_Click);
            zoomToAllMarkersContextMenuItem.Click += new EventHandler(zoomToAllMarkersContextMenuItem_Click);
            showHideLabelsContextMenuItem.Click += new EventHandler(showHideLabelsContextMenuItem_Click);
            zoomToHereContextMenuItem.Click += new EventHandler(zoomToHereContextMenuItem_Click);
            zoomOutContextMenuItem.Click += new EventHandler(zoomOutContextMenuItem_Click);
            streetViewOverlayContextMenuItem.Click += new EventHandler(streetViewOverlayContextMenuItem_Click);
            streetViewContextMenuItem.Click += new EventHandler(streetViewContextMenuItem_Click);

            editMarkerContextMenuItem.Click += new EventHandler(editMarkerContextMenuItem_Click);
            deleteMarkerContextMenuItem.Click += new EventHandler(deleteMarkerContextMenuItem_Click);
            markerZoomToContextMenuItem.Click += new EventHandler(zoomToHereContextMenuItem_Click);
            markerZoomOutContextMenuItem.Click += new EventHandler(zoomOutContextMenuItem_Click);
            markerStreetViewOverlayContextMenuItem.Click += new EventHandler(streetViewOverlayContextMenuItem_Click);
            markerStreetViewContextMenuItem.Click += new EventHandler(streetViewContextMenuItem_Click);
        }



        #endregion Constructors

        #region Properties

        [DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
        public string HelpNamespace
        {
            get { return _helpNamespace; }
            set { _helpNamespace = value; }
        }

        [DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
        public string HelpTopic
        {
            get { return _helpTopic; }
            set { _helpTopic = value; }
        }

        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public Map Map
        {
            get { return _map; }
            set
            {
                if (_map != value)
                {
                    _map = value;
                    if (_map != null)
                    {
                        _map.MapLocations.RaiseListChangedEvents = true;
                        _map.MapLocations.ListChanged += new ListChangedEventHandler(MapLocations_ListChanged);
                        _map.MapLocationsChanged += new Map.MapLocationsChangedEventHandler(_map_MapLocationsChanged);
                        DrawMapLocationCollection(_map.MapLocations);
                    }
                }
            }
        }

        public MapLocation InitialCenterAndZoom
        {
            set
            {
                _initialCenterAndZoom = value;
            }
        }

        public int InitialMapType
        {
            set { _initialMapType = value; }
        }

        public MapLocation SelectedMarker
        {
            get { return _selectedMarker; }
            set { _selectedMarker = value; }
        }

        public ZoomHistory ZoomHistory
        {
            get { return _zoomHistory; }
            set { _zoomHistory = value; }
        }

        //public MapLocation LastCenter
        //{
        //    get { return _lastCenter; }
        //    set { _lastCenter = value; }    // Setter shld only be used by the control.
        //}

        public MapLocation LastCenter
        {
            get
            {
                try
                {
                    return _zoomHistory.Peek();
                }
                catch
                {
                    return null;
                }
            }
        }

        #endregion // Properties

        #region Event Handlers

        void _map_MapLocationsChanged(object sender, EventArgs e)
        {
            MapLocationCollection markers = sender as MapLocationCollection;
            if (e is ListChangedEventArgs)
            {
                ListChangedEventArgs ev = e as ListChangedEventArgs;
                switch (ev.ListChangedType)
                {
                    case ListChangedType.ItemAdded:
                        DrawMapLocationCollection(markers);
                        break;
                    case ListChangedType.ItemDeleted:
                        DeleteMarker(markers.MostRecentlyRemoved);
                        break;
                    case ListChangedType.ItemChanged:
                        MapLocation loc = markers[ev.OldIndex];

                        switch (ev.PropertyDescriptor.Name)
                        {
                            case "Label":
                            case "Lat":
                            case "Lng":
                                SetMarkerProperties(loc);
                                break;
                        }

                        break;
                }
            }
            else
            {
                DrawMapLocationCollection(markers);
            }
        }

        //void Document_MouseDown(object sender, HtmlElementEventArgs e)
        //{
        //    MessageBox.Show("down" + ", x: " + e.MousePosition.X + ", Y: " + e.MousePosition.Y);
        //}

        void MapLocations_ListChanged(object sender, ListChangedEventArgs e)
        {
            DrawMapLocationCollection((MapLocationCollection)sender);
        }

        private void DrawMapLocationCollection(MapLocationCollection locations)
        {
            if (locations == null) return;
            foreach (MapLocation loc in locations)
            {
                if (loc.Id == 0)
                {
                    loc.Id = NewMarker(loc);
                }
            }

        }

        private void newMarkerContextMenuItem_Click(object sender, EventArgs e)
        {
            MapLocation newMarker = new MapLocation("",
                    _rightClickMapLocation.Lat,
                    _rightClickMapLocation.Lng);
            newMarker.Label = "";
            newMarker.Id = NewMarker(newMarker);
            Map.MapLocations.Add(newMarker);

            FrmEditMarker frm = new FrmEditMarker();
            frm.Marker = newMarker;
            frm.Location = RightClickMousePosition;
            frm.ShowDialog();
        }

        private void zoomToAllMarkersContextMenuItem_Click(object sender, EventArgs e)
        {
            ZoomToAllMarkers();
        }

        bool _showHideLabelsStateIsHide = true;
        void showHideLabelsContextMenuItem_Click(object sender, EventArgs e)
        {
            ToolStripMenuItem menuItem = sender as ToolStripMenuItem;
            if (menuItem == null) return;
            _showHideLabelsStateIsHide = !_showHideLabelsStateIsHide;
            if (_showHideLabelsStateIsHide)
            {
                menuItem.Text = "Show All Labels";
                HideAllLabels();
            }
            else
            {
                menuItem.Text = "Hide All Labels";
                ShowAllLabels();
            }
        }

        void zoomToHereContextMenuItem_Click(object sender, EventArgs e)
        {
            MapLocation loc = new MapLocation("",
                    _rightClickMapLocation.Lat,
                    _rightClickMapLocation.Lng,
                    _rightClickMapLocation.ZoomLevel);
            SetCenterAndZoom(loc);
            ZoomIn();
        }

        void zoomOutContextMenuItem_Click(object sender, EventArgs e)
        {
            ZoomOut();
        }




        private void deleteMarkerContextMenuItem_Click(object sender, EventArgs e)
        {
            Map.MapLocations.Remove(SelectedMarker);
            //DeleteMarker(SelectedMarker); // Deletes it from the web control
        }

        private void editMarkerContextMenuItem_Click(object sender, EventArgs e)
        {
            FrmEditMarker frm = new FrmEditMarker();
            frm.Marker = SelectedMarker;
            frm.Location = RightClickMousePosition;
            frm.ShowDialog();
            //MessageBox.Show("Editting Marker " + SelectedMarker.Id);
        }

        bool _streetViewOverlayIsVisible = false;
        void streetViewOverlayContextMenuItem_Click(object sender, EventArgs e)
        {
            StreetViewOverlay(!_streetViewOverlayIsVisible);
            _streetViewOverlayIsVisible = !_streetViewOverlayIsVisible;
            if (_streetViewOverlayIsVisible)
            {
                streetViewOverlayContextMenuItem.Text = "Hide Street View Overlay";
                markerStreetViewOverlayContextMenuItem.Text = "Hide Street View Overlay";
            }
            else
            {
                streetViewOverlayContextMenuItem.Text = "Show Street View Overlay";
                markerStreetViewOverlayContextMenuItem.Text = "Show Street View Overlay";
            }
        }



        void streetViewContextMenuItem_Click(object sender, EventArgs e)
        {
            MapLocation loc = new MapLocation("",
                    _rightClickMapLocation.Lat,
                    _rightClickMapLocation.Lng,
                    _rightClickMapLocation.ZoomLevel);

            if (ShowStreetView != null)
            {
                SetStreetViewMarkerLocation(loc);
                ShowStreetView(this, loc);
            }
        }

        private Point _rightClickMousePosition;

        private MapLocation _rightClickMapLocation;

        internal MapLocation RightClickMapLocation
        {
            get { return _rightClickMapLocation; }
            set { _rightClickMapLocation = value; }
        }

        internal Point RightClickMousePosition
        {
            get { return _rightClickMousePosition; }
            set { _rightClickMousePosition = value; }
        }

        void GoogleMapControl_MapClick(GoogleMapControl sender, MapClickEventArgs e)
        {
            //string msg = "Lat: " + e.MapLocation.Lat +
            //        ", Lng: " + e.MapLocation.Lng;
            //AppUi.StatusText = e.Button + " click: " + msg;
            //if (e.Button == MouseButtons.Right)
            //{
            //    _rightClickMousePosition = e.MousePosition;
            //    _rightClickMapLocation = e.MapLocation;
            //    zoomToAllMarkersContextMenuItem.Visible = Map.MapLocations.Count > 0;
            //    nonMarkerContextMenuStrip.Show(sender, e.MousePosition);
            //}
        }

        private void GoogleMapControl_DocumentCompleted(object sender, WebBrowserDocumentCompletedEventArgs e)
        {
            //Document.MouseDown += new HtmlElementEventHandler(Document_MouseDown);
            Document.MouseMove += new HtmlElementEventHandler(Document_MouseMove);
            Document.Click += new HtmlElementEventHandler(Document_Click);
            //Document.MouseUp += new HtmlElementEventHandler(Document_MouseUp);
            Document.MouseOver += new HtmlElementEventHandler(Document_MouseOver);


            // Register the callback handler (for window.external calls in javascript):
            ICustomDoc cDoc = (ICustomDoc)((SHDocVw.WebBrowser)this.ActiveXInstance).Document;
            cDoc.SetUIHandler(new DocHostUIHandler(this));

            this.PreviewKeyDown += new PreviewKeyDownEventHandler(GoogleMapControl_PreviewKeyDown);
            AppUi.StatusText = "Ready";
        }


        // PreviewKeyDown event is fired twice. Microsoft bug. This is workaround - ignore second event
        private bool _skipFirstPreviewKeyDownEvent = false;

        void GoogleMapControl_PreviewKeyDown(object sender, PreviewKeyDownEventArgs e)
        {
            if (_skipFirstPreviewKeyDownEvent)
            {
                _skipFirstPreviewKeyDownEvent = false;
                return;
            }
            else
            {
                _skipFirstPreviewKeyDownEvent = true;
            }

            //AppUi.StatusText = "Key: " + e.Modifiers + " " + e.KeyCode;

            int finePanNrPixels = 100;
            int superFinePanNrPixels = 10;
            switch (e.KeyCode)
            {
                case Keys.Y:
                    if (e.Control) RedoZoom();
                    break;
                case Keys.Z:
                    if (e.Control) UndoZoom();
                    break;
                case Keys.Oemplus:
                case Keys.Add:
                    if (e.Control) ZoomIn();
                    break;
                case Keys.OemMinus:
                case Keys.Subtract:
                    if (e.Control) ZoomOut();
                    break;
                case Keys.Up:
                    if (e.Control && e.Shift) PanByPixels(0, superFinePanNrPixels);
                    else if (e.Control || e.Shift) PanByPixels(0, finePanNrPixels);
                    else PanInDirection(0, 1);
                    break;
                case Keys.Down:
                    if (e.Control && e.Shift) PanByPixels(0, -superFinePanNrPixels);
                    else if (e.Control || e.Shift) PanByPixels(0, -finePanNrPixels);
                    else PanInDirection(0, -1);
                    break;
                case Keys.Left:
                    if (e.Control && e.Shift) PanByPixels(superFinePanNrPixels, 0);
                    else if (e.Control || e.Shift) PanByPixels(finePanNrPixels, 0);
                    else PanInDirection(1, 0);
                    break;
                case Keys.Right:
                    if (e.Control && e.Shift) PanByPixels(-superFinePanNrPixels, 0);
                    else if (e.Control || e.Shift) PanByPixels(-finePanNrPixels, 0);
                    else PanInDirection(-1, 0);
                    break;
                case Keys.F4: // Shut down the parent form
                    if (e.Modifiers == Keys.Alt) ParentForm.Close(); // Shut down the parent form
                    break;
                case Keys.Help:
                case Keys.F1:
                    if (HelpNamespace != null && HelpTopic != null)
                    {
                        Help.ShowHelp(this, HelpNamespace, HelpNavigator.Topic, HelpTopic);
                    }
                    break;
            }
        }


        string _currentHtmlElementId;
        void Document_MouseOver(object sender, HtmlElementEventArgs e)
        {
            //Debug.Print("{0} Document_MouseOver, e.FromElement: {1}, e.ToElement: {2}\n Id:{3},Name:{4}", _i++, e.FromElement, e.ToElement,
            //        e.ToElement.Id, e.ToElement.Name);
            if (e.ToElement != null)
            {
                _currentHtmlElementId = e.ToElement.Id;
            }
            //throw new NotImplementedException();
        }

        void Document_MouseMove(object sender, HtmlElementEventArgs e)
        {
            if (MapMouseMove != null)
            {
                MapLocation loc = PixelToMapLocation(e.MousePosition.X, e.MousePosition.Y);
                MapMouseMove(this, loc);
            }
        }

        void Document_MouseUp(object sender, HtmlElementEventArgs e)
        {
            // Generated by javascript on right-mouse click.
            // Generate a MapClick right-mouse click event:
            if (MapClick != null)
            {
                if (e.MouseButtonsPressed == MouseButtons.Right)
                {
                    MapClickEventArgs args = new MapClickEventArgs();
                    args.MousePosition = e.MousePosition;
                    args.MapLocation = PixelToMapLocation(e.MousePosition.X, e.MousePosition.Y);
                    args.Button = MouseButtons.Right;
                    Debug.Print("_currentHtmlElementId: {0}", _currentHtmlElementId);
                    string markerIds = (string)Document.InvokeScript("getMarkerHtmlElementIds");
                    Debug.Print("markerIds: {0}", markerIds);
                    //args.MarkerId = Convert.ToInt32(Document.InvokeScript("getRmbMarkerId"));
                    //Debug.Print("args.MarkerId: {0}", args.MarkerId);
                    MapClick(this, args);
                }
            }
        }

        void Document_Click(object sender, HtmlElementEventArgs e)
        {
            // Generated by javascript on left-mouse click.
            // Generate a MapClick left-mouse click event:
            if (MapClick != null)
            {
                MapClickEventArgs args = new MapClickEventArgs();
                args.MapLocation = PixelToMapLocation(e.MousePosition.X, e.MousePosition.Y);
                args.Button = MouseButtons.Left;
                MapClick(this, args);
            }
        }

        #endregion // Event Handlers

        #region Public Methods

        /// <summary>
        /// This *must* be called after all the InitialCenterAndZoom, and MapLocations 
        /// have been set, to display the map.
        /// </summary>
        public void DisplayMap()
        {
            // _map.setCenter(new GLatLng(-31.94, 115.83), 3);
            string sSetCenter =
                "_map.setCenter(new GLatLng(" + _initialCenterAndZoom.Lat +
                ", " + _initialCenterAndZoom.Lng +
                "), " + _initialCenterAndZoom.ZoomLevel +
                ");";
            Dictionary<string, string> replacements = new Dictionary<string, string>();
            replacements.Add(INIT_CENTER_AND_ZOOM_TOKEN, sSetCenter);
            string sSetMapType =
                "setMapType(" + _initialMapType + ");";
            replacements.Add(INIT_MAP_TYPE_TOKEN, sSetMapType);
            ////replacements.Add(INIT_MARKERS_TOKEN, InitialMarkersJavaScript());
            AppUtil.WriteResourceToDisk("MapApp.Ui.Controls.Resources.html.map.html",
                mapFilePath, replacements);

            Url = new Uri(mapFilePath);
        }


        public void PanTo(MapLocation loc)
        {
            bool oldShowAllLabelsState = _showAllLabelsState;
            Object[] objArray = { loc.Lat, loc.Lng };

            if (oldShowAllLabelsState) HideAllLabels();
            Document.InvokeScript("panTo", objArray);
            if (oldShowAllLabelsState) ShowAllLabels();
        }

        public void ZoomTo(MapLocation loc)
        {
            bool oldShowAllLabelsState = _showAllLabelsState;
            Object[] objArray = { loc.Lat, loc.Lng, loc.ZoomLevel };

            if (oldShowAllLabelsState) HideAllLabels();
            Document.InvokeScript("zoomTo", objArray);
            if (oldShowAllLabelsState) ShowAllLabels();
        }

        public void ZoomToAllMarkers()
        {
            Document.InvokeScript("zoomToAllMarkers");
        }

        public void SetOverviewMapVisible(bool visible)
        {
            Object[] objArray = { visible };
            Document.InvokeScript("setOverviewMapVisible", objArray);
        }

        public void SetScaleBarVisible(bool visible)
        {
            Object[] objArray = { visible };
            Document.InvokeScript("setScaleBarVisible", objArray);
        }

        public void ShowAllLabels()
        {
            _showAllLabelsState = true;
            Document.InvokeScript("showAllTooltips");
        }

        public void HideAllLabels()
        {
            _showAllLabelsState = false;
            Document.InvokeScript("hideAllTooltips");
        }

        /// <summary>
        /// All this does is add a marker to the Map. Doesn't add it to
        /// MapLocations list. To do that use AddNewMarker(loc)
        /// </summary>
        /// <param name="loc"></param>
        /// <returns></returns>
        public int NewMarker(MapLocation loc)
        {
            string toolTip = string.IsNullOrEmpty(loc.Label) ? "" : loc.Label;
            Object[] objArray = null;
            if (loc.ImageUrl != null && loc.ImageUrl.ToString() != "")
            {
                Object[] a1 = { loc.Lat, loc.Lng, toolTip, loc.ImageUrl.ToString() };
                objArray = a1;
            }
            else
            {
                Object[] a2 = { loc.Lat, loc.Lng, toolTip };
                objArray = a2;
            }
            int id = Convert.ToInt32(Document.InvokeScript("newMarker", objArray));
            return id;
        }

        /// <summary>
        /// Adds a new marker to the map, and adds the marker to the MapLocations
        /// collection.
        /// </summary>
        /// <param name="loc"></param>
        public void AddNewMarker(MapLocation loc)
        {
            loc.Id = NewMarker(loc);
            Map.MapLocations.Add(loc);
        }

        public void SetMarkerProperties(MapLocation loc)
        {
            string toolTip = string.IsNullOrEmpty(loc.Label) ? "" : loc.Label;
            Object[] objArray = { loc.Id, loc.Lat, loc.Lng, toolTip };
            Document.InvokeScript("setMarkerProperties", objArray);
        }

        public MapLocation GetCenterAndZoom()
        {
            string s = (string)Document.InvokeScript("getCenterAndZoom");
            double lat = 0.0, lng = 0.0;
            int zoom = 0;
            try
            {
                if (s != null)
                {
                    string[] sArray = s.Split(':');
                    lat = Convert.ToDouble(sArray[0]);
                    lng = Convert.ToDouble(sArray[1]);
                    zoom = Convert.ToInt32(sArray[2]);
                }
            }
            catch { } // ignore exceptions
            MapLocation result = new MapLocation("", lat, lng, zoom);
            return result;
        }

        public void SetCenterAndZoom(MapLocation loc)
        {
            Object[] objArray = { loc.Lat, loc.Lng, loc.ZoomLevel };
            Document.InvokeScript("setCenterAndZoom", objArray);
        }

        public void SetCenterAndZoomNoHistory(MapLocation loc)
        {
            Object[] objArray = { loc.Lat, loc.Lng, loc.ZoomLevel };
            Document.InvokeScript("setCenterAndZoomNoHistory", objArray);
        }

        public void StreetViewOverlay(bool show)
        {
            if (show)
                Document.InvokeScript("showStreetViewOverlay");
            else 
                Document.InvokeScript("hideStreetViewOverlay");
        }
        
        public void SetStreetViewMarkerLocation(MapLocation loc)
        {
            Object[] objArray = { loc.Lat, loc.Lng };
            Document.InvokeScript("setStreetViewMarkerLocation", objArray);
        }

        public int GetMapType()
        {
            int mapType = (int)Document.InvokeScript("getMapType");
            return Convert.ToInt32(mapType);
        }

        public void SetMapType(int mapType)
        {
            Object[] objArray = { mapType };
            Document.InvokeScript("setMapType", objArray);
        }

        public void UndoZoom()
        {
            SetCenterAndZoomNoHistory(ZoomHistory.Pop());
        }

        public void RedoZoom()
        {
            SetCenterAndZoomNoHistory(ZoomHistory.UnPop());
        }

        public void ZoomIn()
        {
            Document.InvokeScript("zoomIn");
        }

        public void ZoomOut()
        {
            Document.InvokeScript("zoomOut");
        }

        /// <summary>
        /// Starts a pan animation by half the width of the map 
        /// in the indicated directions. +1 is right and down, 
        /// -1 is left and up, respectively. 
        /// </summary>
        /// <param name="dx">one of -1 (left) 0 +1 (right)</param>
        /// <param name="dy">one of -1 (up) 0 +1 (down)</param>
        public void PanInDirection(int dx, int dy)
        {
            Object[] objArray = { dx, dy };
            Document.InvokeScript("panInDirection", objArray);
        }

        public void PanByPixels(int pixelsX, int pixelsY)
        {
            Object[] objArray = { pixelsX, pixelsY };
            Document.InvokeScript("panByPixels", objArray);
        }

        public void RaiseShowStreetView(MapLocation loc)
        {
            if (ShowStreetView != null)
            {
                ShowStreetView(this, loc);
            }
        }

        #endregion Public Methods

        private void DeleteMarker(MapLocation loc)
        {
            Object[] objArray = { loc.Id };
            Document.InvokeScript("deleteMarker", objArray);
        }

        internal int GetCurrentZoomLevel()
        {
            int zoomLevel = 0;
            try
            {
                zoomLevel = (int)Document.InvokeScript("getCurrentZoomLevel");
            }
            catch { } // map may not yet be fully created
            return zoomLevel;
        }

        internal MapLocation PixelToMapLocation(int pixelX, int pixelY)
        {
            Object[] objArray = { pixelX, pixelY };
            string llString = (string)Document.InvokeScript("pixelToLatLng", objArray);
            double lat = 0.0, lng = 0.0;
            if (llString != null)
            {
                string[] llArray = llString.Split(':');
                try
                {
                    lat = Convert.ToDouble(llArray[0]);
                    lng = Convert.ToDouble(llArray[1]);
                }
                catch { } // ignore exceptions
            }
            MapLocation result = new MapLocation("", lat, lng, GetCurrentZoomLevel());
            return result;
        }

        private Form ParentForm
        {
            get
            {
                for (Control parent = this.Parent; parent != null; parent = parent.Parent)
                {
                    if (parent is Form) return parent as Form;
                }
                return null;
            }
        }

        //private MapLocation _centerBeforeResize;
        private FormWindowState _windowStateBeforeResize;

        private void GoogleMapControl_Resize(object sender, EventArgs e)
        {
            if (ParentForm == null) return;
            switch (ParentForm.WindowState)
            {
                case FormWindowState.Maximized:
                case FormWindowState.Normal:
                    if (_windowStateBeforeResize != ParentForm.WindowState &&
                        _windowStateBeforeResize != FormWindowState.Minimized &&
                        LastCenter != null)
                    {
                        //ZoomHistory.Pop();
                        //SetCenterAndZoom(ZoomHistory.Pop());
                    }
                    break;
            }
            // set old values for next time:
            //try
            //{
            //    //_centerBeforeResize = GetCenterAndZoom();
            //    AppUi.StatusText = "Center before resize: " + _centerBeforeResize;
            //}
            //catch { } // ignore exception - MapControl might not yet exist
            _windowStateBeforeResize = ParentForm.WindowState;
        }

    }


    public class MapClickEventArgs : EventArgs
    {
        public MapLocation MapLocation;
        public Point MousePosition;
        public MouseButtons Button;
        public int MarkerId;
    }

    public class MapExternal : IMapExternalInterface
    {
        //private Map _map;
        private GoogleMapControl _mapControl;
        public MapExternal(GoogleMapControl mapControl)
        {
            _mapControl = mapControl;
        }

        #region IMapExternalInterface Members

        public void pushZoomHistory(double lat, double lng, int zoomLevel)
        {
            _mapControl.ZoomHistory.Push(new MapLocation("", lat, lng, zoomLevel));
            //AppUi.StatusText = "Zoom: " + lat + ", " + lng + ", " + zoomLevel;
        }

        public void addAllMarkers()
        {
            if (_mapControl.Map.MapLocations != null)
            {
                foreach (MapLocation loc in _mapControl.Map.MapLocations)
                {
                    loc.Id = _mapControl.NewMarker(loc);
                }
            }
        }

        public void markerMoved(int markerId, double toLat, double toLng)
        {
            if (_mapControl.Map.MapLocations != null)
            {
                // TODO: Need to implement a dictionary of locationId to location for lookup
                MapLocation loc = _mapControl.Map.MapLocations.Find(markerId);
                if (loc != null)
                {
                    loc.Lat = toLat;
                    loc.Lng = toLng;
                }
            }
        }

        public void streetViewMarkerDragEndHandler(double lat, double lng)
        {
            MapLocation newStreetViewMarkerLocation = new MapLocation("", lat, lng);
            _mapControl.RaiseShowStreetView(newStreetViewMarkerLocation);
        }


        public void nonMarkerContextMenu(int pixelX, int pixelY)
        {
            Point mousePosition = new Point(pixelX, pixelY);
            _mapControl.RightClickMousePosition = mousePosition;
            //_rightClickMapLocation = e.MapLocation;
            int currentZoomLevel = _mapControl.GetCurrentZoomLevel();
            MapLocation rightClickLocation = _mapControl.PixelToMapLocation(pixelX, pixelY);
            _mapControl.RightClickMapLocation = new MapLocation("", rightClickLocation.Lat,
                    rightClickLocation.Lng, currentZoomLevel);
            if (_mapControl.Map.MapLocations.Count > 0)
            {
                _mapControl.zoomToAllMarkersContextMenuItem.Visible = true;
                _mapControl.showHideLabelsContextMenuItem.Visible = true;
            }
            else
            {
                _mapControl.zoomToAllMarkersContextMenuItem.Visible = false;
                _mapControl.showHideLabelsContextMenuItem.Visible = false;

            }

            _mapControl.nonMarkerContextMenuStrip.Show(_mapControl, mousePosition);

            //throw new NotImplementedException();
        }


        public void markerContextMenu(int markerId, int pixelX, int pixelY)
        {
            _mapControl.RightClickMousePosition = new Point(pixelX, pixelY);
            //_rightClickMapLocation = e.MapLocation;
            int currentZoomLevel = _mapControl.GetCurrentZoomLevel();
            MapLocation rightClickLocation = _mapControl.PixelToMapLocation(pixelX, pixelY);
            _mapControl.RightClickMapLocation = new MapLocation("", rightClickLocation.Lat,
                    rightClickLocation.Lng, currentZoomLevel);
            _mapControl.SelectedMarker = _mapControl.Map.MapLocations.Find(markerId);
            _mapControl.markerContextMenuStrip.Show(_mapControl, new Point(pixelX, pixelY));

            //throw new NotImplementedException();
        }

        #endregion
    }

    [InterfaceType(ComInterfaceType.InterfaceIsIDispatch)]
    public interface IMapExternalInterface
    {
        [DispId(0)]
        void addAllMarkers();

        [DispId(1)]
        void markerMoved(int markerId, double toLat, double toLng);

        [DispId(2)]
        void nonMarkerContextMenu(int pixelX, int pixelY);

        [DispId(3)]
        void markerContextMenu(int markerId, int pixelX, int pixelY);

        [DispId(4)]
        void pushZoomHistory(double lat, double lng, int zoomLevel);

        [DispId(5)]
        void streetViewMarkerDragEndHandler(double lat, double lng);
    }

    public class DocHostUIHandler : IDocHostUIHandler
    {
        private GoogleMapControl _mapControl;
        public DocHostUIHandler(GoogleMapControl mapControl)
        {
            _mapControl = mapControl;
        }

        // IDocHostUIHandler implementation

        void IDocHostUIHandler.EnableModeless(int fEnable)
        {

        }

        void IDocHostUIHandler.FilterDataObject(MsHtmHstInterop.IDataObject pDO, out MsHtmHstInterop.IDataObject ppDORet)
        {
            ppDORet = null;
        }

        void IDocHostUIHandler.GetDropTarget(MsHtmHstInterop.IDropTarget pDropTarget, out MsHtmHstInterop.IDropTarget ppDropTarget)
        {
            ppDropTarget = null;
        }

        void IDocHostUIHandler.GetExternal(out object ppDispatch)
        {
            ppDispatch = new MapExternal(_mapControl);
        }

        void IDocHostUIHandler.GetHostInfo(ref _DOCHOSTUIINFO pInfo)
        {

        }

        void IDocHostUIHandler.GetOptionKeyPath(out string pchKey, uint dw)
        {
            pchKey = null;
        }

        void IDocHostUIHandler.HideUI()
        {

        }

        void IDocHostUIHandler.OnDocWindowActivate(int fActivate)
        {

        }

        void IDocHostUIHandler.OnFrameWindowActivate(int fActivate)
        {

        }

        void IDocHostUIHandler.ResizeBorder(ref MsHtmHstInterop.tagRECT prcBorder, IOleInPlaceUIWindow pUIWindow, int fRameWindow)
        {

        }

        void IDocHostUIHandler.ShowContextMenu(uint dwID, ref MsHtmHstInterop.tagPOINT ppt, object pcmdtReserved, object pdispReserved)
        {

        }

        void IDocHostUIHandler.ShowUI(uint dwID, IOleInPlaceActiveObject pActiveObject, IOleCommandTarget pCommandTarget, IOleInPlaceFrame pFrame, IOleInPlaceUIWindow pDoc)
        {

        }

        void IDocHostUIHandler.TranslateAccelerator(ref tagMSG lpmsg, ref Guid pguidCmdGroup, uint nCmdID)
        {

        }

        void IDocHostUIHandler.TranslateUrl(uint dwTranslate, ref ushort pchURLIn, IntPtr ppchURLOut)
        {

        }

        void IDocHostUIHandler.UpdateUI()
        {

        }
    }

}
